/* edit-recipients.c - Copy an encrypted file, changing its recipients
 * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
 *               2006, 2007, 2008, 2009, 2010 Free Software Foundation, Inc.
 * Copyright (C) 2010 Nicolas Boullis <nboullis@debian.org>
 *
 * This file is part of gpg-edit-recipients; big parts of it were copied from GnuPG.
 *
 * gpg-edit-recipients is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * gpg-edit-recipients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "gpg.h"
#include "packet.h"
#include "options.h"
#include "status.h"
#include "i18n.h"
#include "pkglue.h"

typedef struct pubkey_list *PUBKEY_LIST;
struct pubkey_list {
    PUBKEY_LIST next;
    PKT_pubkey_enc *packet;
};

typedef struct symkey_list *SYMKEY_LIST;
struct symkey_list {
    SYMKEY_LIST next;
    PKT_symkey_enc *packet;
};

typedef struct mainproc_context *CTX;
struct mainproc_context
{
  struct mainproc_context *anchor;  /* May be useful in the future. */
  PUBKEY_LIST pubkeys;
  SYMKEY_LIST symkeys;
  PKT_public_key *last_pubkey;
  PKT_secret_key *last_seckey;
  PKT_user_id     *last_user_id;
  md_filter_context_t mfx;
    
  /* Name of the file with the complete signature or the file with the
     detached signature.  This is currently only used to deduce the
     file name of the data file if that has not been given. */
  const char *sigfilename;
  
  /* A structure to describe the signed data in case of a detached
     signature. */
  struct 
  {
    /* A file descriptor of the the signed data.  Only used if not -1. */
    int data_fd;
    /* A list of filenames with the data files or NULL. This is only
       used if DATA_FD is -1. */
    strlist_t data_names;
    /* Flag to indicated that either one of the next previous fieldss
       is used.  This is only needed for better readability. */
    int used;
  } signed_data;
  
  DEK *dek;
  int last_was_session_key;
  KBNODE list;      /* The current list of packets. */
  int have_data;
  IOBUF iobuf;      /* Used to get the filename etc. */
  IOBUF out;
  int trustletter;  /* Temporary usage in list_node. */
  struct kidlist_item *pkenc_list; /* List of encryption packets. */
  int any_sig_seen;  /* Set to true if a signature packet has been seen. */
  PK_LIST pk_list;
};

/****************
 * Write pubkey-enc packets from the list of PKs to OUT.
 */
static int
write_pubkey_enc_from_list( PK_LIST pk_list, DEK *dek, IOBUF out )
{
    PACKET pkt;
    PKT_public_key *pk;
    PKT_pubkey_enc  *enc;
    int rc;

    for( ; pk_list; pk_list = pk_list->next ) {
	gcry_mpi_t frame;

	pk = pk_list->pk;

	print_pubkey_algo_note( pk->pubkey_algo );
	enc = xmalloc_clear( sizeof *enc );
	enc->pubkey_algo = pk->pubkey_algo;
	keyid_from_pk( pk, enc->keyid );
	enc->throw_keyid = (opt.throw_keyid || (pk_list->flags&1));

	if(opt.throw_keyid && (PGP2 || PGP6 || PGP7 || PGP8))
	  {
	    log_info(_("you may not use %s while in %s mode\n"),
		     "--throw-keyid",compliance_option_string());
	    compliance_failure();
	  }

	/* Okay, what's going on: We have the session key somewhere in
	 * the structure DEK and want to encode this session key in
	 * an integer value of n bits.	pubkey_nbits gives us the
	 * number of bits we have to use.  We then encode the session
	 * key in some way and we get it back in the big intger value
	 * FRAME.  Then we use FRAME, the public key PK->PKEY and the
	 * algorithm number PK->PUBKEY_ALGO and pass it to pubkey_encrypt
	 * which returns the encrypted value in the array ENC->DATA.
	 * This array has a size which depends on the used algorithm
	 * (e.g. 2 for Elgamal).  We don't need frame anymore because we
	 * have everything now in enc->data which is the passed to
	 * build_packet()
	 */
	frame = encode_session_key (dek, pubkey_nbits (pk->pubkey_algo,
                                                       pk->pkey) );
	rc = pk_encrypt (pk->pubkey_algo, enc->data, frame, pk->pkey);
	gcry_mpi_release (frame);
	if( rc )
	    log_error ("pubkey_encrypt failed: %s\n", gpg_strerror (rc) );
	else {
	    if( opt.verbose ) {
		char *ustr = get_user_id_string_native (enc->keyid);
		log_info(_("%s/%s encrypted for: \"%s\"\n"),
                         gcry_pk_algo_name (enc->pubkey_algo),
                         openpgp_cipher_algo_name (dek->algo),
                         ustr );
		xfree(ustr);
	    }
	    /* and write it */
	    init_packet(&pkt);
	    pkt.pkttype = PKT_PUBKEY_ENC;
	    pkt.pkt.pubkey_enc = enc;
	    rc = build_packet( out, &pkt );
	    if( rc )
	       log_error("build_packet(pubkey_enc) failed: %s\n", g10_errstr(rc));
	}
	free_pubkey_enc(enc);
	if( rc )
	    return rc;
    }
    return 0;
}

static int
symkey_decrypt_seskey( DEK *dek, byte *seskey, size_t slen )
{
  gcry_cipher_hd_t hd;
  byte *data;

  if(slen < 17 || slen > 33)
    {
      log_error ( _("weird size for an encrypted session key (%d)\n"),
		  (int)slen);
      return G10ERR_BAD_KEY;
    }

  if (openpgp_cipher_open (&hd, dek->algo, GCRY_CIPHER_MODE_CFB, 1))
      BUG ();
  if (gcry_cipher_setkey ( hd, dek->key, dek->keylen ))
    BUG ();

  data = xmalloc_clear( slen );
  memcpy(data, seskey, slen);

  gcry_cipher_setiv ( hd, NULL, 0 );
  gcry_cipher_decrypt ( hd, data, slen, NULL, 0 );
  gcry_cipher_close ( hd );

  /* Now we replace the dek components with the real session key to
     decrypt the contents of the sequencing packet. */

  dek->keylen=slen-1;
  dek->algo=data[0];

  if(dek->keylen > DIM(dek->key))
    BUG ();

  /* This is not completely accurate, since a bad passphrase may have
     resulted in a garbage algorithm byte, but it's close enough since
     a bogus byte here will fail later. */
  if(dek->algo==CIPHER_ALGO_IDEA)
    idea_cipher_warn(0);

  memcpy(dek->key, data + 1, dek->keylen);

  xfree(data);

  /*log_hexdump( "thekey", dek->key, dek->keylen );*/

  return 0;
}   

static void
proc_symkey_enc( CTX c, PACKET *pkt )
{
    PKT_symkey_enc *enc;

    enc = pkt->pkt.symkey_enc;
    if (!enc)
        log_error ("invalid symkey encrypted packet\n");
    else if(!c->dek)
      {
        int algo = enc->cipher_algo;
	const char *s = openpgp_cipher_algo_name (algo);

	if (!openpgp_cipher_test_algo (algo))
	  {
	    if(!opt.quiet)
	      {
		if(enc->seskeylen)
		  log_info(_("%s encrypted session key\n"), s );
		else
		  log_info(_("%s encrypted data\n"), s );
	      }
	  }
	else
	  log_error(_("encrypted with unknown algorithm %d\n"), algo );

	if(openpgp_md_test_algo (enc->s2k.hash_algo))
	  {
	    log_error(_("passphrase generated with unknown digest"
			" algorithm %d\n"),enc->s2k.hash_algo);
	    s=NULL;
	  }

	c->last_was_session_key = 2;
	if(!s || opt.list_only)
	  goto leave;

	if(opt.override_session_key)
	  {
	    c->dek = xmalloc_clear( sizeof *c->dek );
	    if(get_override_session_key(c->dek, opt.override_session_key))
	      {
		xfree(c->dek);
		c->dek = NULL;
	      }
	  }
	else
	  {
	    c->dek = passphrase_to_dek (NULL, 0, algo, &enc->s2k, 3,
                                        NULL, NULL);
	    if(c->dek)
	      {
		c->dek->symmetric=1;

		/* FIXME: This doesn't work perfectly if a symmetric
		   key comes before a public key in the message - if
		   the user doesn't know the passphrase, then there is
		   a chance that the "decrypted" algorithm will happen
		   to be a valid one, which will make the returned dek
		   appear valid, so we won't try any public keys that
		   come later. */
		if(enc->seskeylen)
		  {
		    if(symkey_decrypt_seskey(c->dek, enc->seskey,
					     enc->seskeylen))
		      {
			xfree(c->dek);
			c->dek=NULL;
		      }
		  }
		else
		  c->dek->algo_info_printed = 1;
	      }
	  }
      }

 leave:
    {
	SYMKEY_LIST x = xmalloc( sizeof *x);
	x->packet = enc;
	x->next = c->symkeys;
	c->symkeys = x;
	init_packet(pkt);
    }
    /* c->symkeys++; */
    free_packet(pkt);
}

static void
proc_pubkey_enc( CTX c, PACKET *pkt )
{
    PKT_pubkey_enc *enc;
    int result = 0;

    /* check whether the secret key is available and store in this case */
    enc = pkt->pkt.pubkey_enc;
    /*printf("enc: encrypted by a pubkey with keyid %08lX\n", enc->keyid[1] );*/
    /* Hmmm: why do I have this algo check here - anyway there is
     * function to check it. */
    if( opt.verbose )
	log_info(_("public key is %s\n"), keystr(enc->keyid) );

    if( is_status_enabled() ) {
	char buf[50];
	sprintf(buf, "%08lX%08lX %d 0",
		(ulong)enc->keyid[0], (ulong)enc->keyid[1], enc->pubkey_algo );
	write_status_text( STATUS_ENC_TO, buf );
    }

    if( !opt.list_only && opt.override_session_key ) {
	/* It does not make much sense to store the session key in
	 * secure memory because it has already been passed on the
	 * command line and the GCHQ knows about it.  */
	c->dek = xmalloc_clear( sizeof *c->dek );
	result = get_override_session_key ( c->dek, opt.override_session_key );
	if ( result ) {
	    xfree(c->dek); c->dek = NULL;
	}
    }
    else if( is_ELGAMAL(enc->pubkey_algo)
             || enc->pubkey_algo == PUBKEY_ALGO_DSA
             || is_RSA(enc->pubkey_algo)
             || enc->pubkey_algo == PUBKEY_ALGO_ELGAMAL) {
      /* Note that we also allow type 20 Elgamal keys for decryption.
         There are still a couple of those keys in active use as a
         subkey.  */

      /* FIXME: Store this all in a list and process it later so that
         we can prioritize what key to use.  This gives a better user
         experience if wildcard keyids are used.  */
	if ( !c->dek && ((!enc->keyid[0] && !enc->keyid[1])
                          || opt.try_all_secrets
			  || !seckey_available( enc->keyid )) ) {
	    if( opt.list_only )
		result = -1;
	    else {
		c->dek = xmalloc_secure_clear( sizeof *c->dek );
		if( (result = get_session_key( enc, c->dek )) ) {
		    /* error: delete the DEK */
		    xfree(c->dek); c->dek = NULL;
		}
	    }
	}
	else
	    result = G10ERR_NO_SECKEY;
    }
    else
	result = G10ERR_PUBKEY_ALGO;

    if( result == -1 )
	;
    else
      {
	PUBKEY_LIST x = xmalloc( sizeof *x);
	x->packet = enc;
	x->next = c->pubkeys;
	c->pubkeys = x;
	init_packet(pkt);
        /* store it for later display */
	/*
	struct kidlist_item *x = xmalloc( sizeof *x );
	x->kid[0] = enc->keyid[0];
	x->kid[1] = enc->keyid[1];
	x->pubkey_algo = enc->pubkey_algo;
	x->reason = result;
	x->next = c->pkenc_list;
	c->pkenc_list = x;
	*/

        if( !result && opt.verbose > 1 )
	  log_info( _("public key encrypted data: good DEK\n") );
      }

    free_packet(pkt);
}

static int
copy_packet( IOBUF inp, IOBUF out, int pkttype,
	     unsigned long pktlen, int partial )
{
    int rc;
    int n;
    char buf[100];

    if( partial ) {
	while( (n = iobuf_read( inp, buf, 100 )) != -1 )
	    if( (rc=iobuf_write(out, buf, n )) )
		return rc; /* write error */
    }
    else if( !pktlen && pkttype == PKT_COMPRESSED ) {
	log_debug("copy_packet: compressed!\n");
	/* compressed packet, copy till EOF */
	while( (n = iobuf_read( inp, buf, 100 )) != -1 )
	    if( (rc=iobuf_write(out, buf, n )) )
		return rc; /* write error */
    }
    else {
	for( ; pktlen; pktlen -= n ) {
	    n = pktlen > 100 ? 100 : pktlen;
	    n = iobuf_read( inp, buf, n );
	    if( n == -1 )
		return gpg_error (GPG_ERR_EOF);
	    if( (rc=iobuf_write(out, buf, n )) )
		return rc; /* write error */
	}
    }
    return 0;
}


static void
proc_encrypted( CTX c, PACKET *pkt )
{
    int rc;
    if ( c->pk_list && !c->dek)
      {
	return;
      }
    {
      PUBKEY_LIST x;
      for (x = c->pubkeys; x; x = x->next) {
	  PACKET pkt2;
	  init_packet(&pkt2);
	  pkt2.pkttype = PKT_PUBKEY_ENC;
	  pkt2.pkt.pubkey_enc = x->packet;
	  rc = build_packet( c->out, &pkt2 );
	  if( rc )
	      log_error("build_packet(pubkey_enc) failed: %s\n", g10_errstr(rc));
      }
    }
    write_pubkey_enc_from_list( c->pk_list, c->dek, c->out );
    {
      SYMKEY_LIST x;
      for (x = c->symkeys; x; x = x->next) {
	  PACKET pkt2;
	  init_packet(&pkt2);
	  pkt2.pkttype = PKT_SYMKEY_ENC;
	  pkt2.pkt.symkey_enc = x->packet;
	  rc = build_packet( c->out, &pkt2 );
	  if( rc )
	      log_error("build_packet(pubkey_enc) failed: %s\n", g10_errstr(rc));
      }
    }
    if (pkt->pkttype == PKT_ENCRYPTED_MDC) {
      if (pkt->pkt.encrypted->len)
        pkt->pkt.encrypted->len -= 23;
    }
    build_packet( c->out, pkt );
    if (pkt->pkttype == PKT_ENCRYPTED_MDC) {
      if (pkt->pkt.encrypted->len)
        pkt->pkt.encrypted->len += 22;
    }
    copy_packet( c->iobuf, c->out, pkt->pkttype,
		 pkt->pkt.encrypted->len, pkt->pkt.encrypted->is_partial );
    while (c->pubkeys) {
        PUBKEY_LIST x = c->pubkeys;
	c->pubkeys = x->next;
        free_pubkey_enc(x->packet);
	xfree(x);
    }
    while (c->symkeys) {
        SYMKEY_LIST x = c->symkeys;
	c->symkeys = x->next;
        free_symkey_enc(x->packet);
	xfree(x);
    }
    xfree(c->dek); c->dek = NULL;
    free_packet(pkt);
}

int
edit_recipients( const char *filename, strlist_t new_recipients )
{
    IOBUF fp;
    armor_filter_context_t *afx = NULL;
    progress_filter_context_t *pfx;
    int rc;
    int no_out = 0;
    PK_LIST pk_list;

    pfx = new_progress_context ();

    if( (rc=build_pk_list( new_recipients, &pk_list, PUBKEY_USAGE_ENC)) )
      {
        release_progress_context (pfx);
	return rc;
      }

    /* Open the message file.  */
    fp = iobuf_open(filename);
    if (fp && is_secured_file (iobuf_get_fd (fp)))
      {
        iobuf_close (fp);
        fp = NULL;
        errno = EPERM;
      }
    if( !fp ) {
        rc = gpg_error_from_syserror ();
	log_error (_("can't open `%s': %s\n"), print_fname_stdin(filename),
                   gpg_strerror (rc));
        release_progress_context (pfx);
	return rc;
    }

    handle_progress (pfx, fp, filename);

    if( !opt.no_armor ) {
	if( use_armor_filter( fp ) ) {
            afx = new_armor_context ();
	    push_armor_filter ( afx, fp );
	}
    }

    if( !opt.outfile ) {
	no_out = 1;
	opt.outfile = "-";
    }

    {
        CTX c = xmalloc_clear( sizeof *c );
	PACKET *pkt = xmalloc( sizeof *pkt );
	int any_data=0;
	int newpkt;

	if( (rc = open_outfile( opt.outfile, opt.armor? 1:0, &c->out )) )
	  goto leave;

	if ( opt.armor )
	  {
	    afx = new_armor_context ();
	    push_armor_filter (afx, c->out);
	  }

	c->iobuf = fp;
	c->pk_list = pk_list;

	init_packet(pkt);
	while( (rc=parse_packet(fp, pkt)) != -1 ) {
	    any_data = 1;
	    if( rc ) {
	        free_packet(pkt);
		/* stop processing when an invalid packet has been encountered
		 * but don't do so when we are doing a --list-packets. */
		if (gpg_err_code (rc) == GPG_ERR_INV_PACKET
		    && opt.list_packets != 2 )
		    break;
		continue;
	    }
	    newpkt = -1;
	    switch( pkt->pkttype ) {
	      case PKT_PUBLIC_KEY:
	      case PKT_SECRET_KEY:
	      case PKT_USER_ID:
	      case PKT_SIGNATURE:
	      case PKT_PLAINTEXT:
	      case PKT_COMPRESSED:
	      case PKT_ONEPASS_SIG:
	      case PKT_GPG_CONTROL:
                write_status_text( STATUS_UNEXPECTED, "0" );
		rc = G10ERR_UNEXPECTED;
		goto leave;
	      case PKT_SYMKEY_ENC:  proc_symkey_enc( c, pkt ); break;
	      case PKT_PUBKEY_ENC:  proc_pubkey_enc( c, pkt ); break;
	      case PKT_ENCRYPTED:
	      case PKT_ENCRYPTED_MDC: proc_encrypted( c, pkt ); break;
	      default: newpkt = 0; break;
	    }
	    /* This is a very ugly construct and frankly, I don't remember why
	     * I used it.  Adding the MDC check here is a hack.
	     * The right solution is to initiate another context for encrypted
	     * packet and not to reuse the current one ...  It works right
	     * when there is a compression packet inbetween which adds just
	     * an extra layer.
	     * Hmmm: Rewrite this whole module here?? 
	     */
	    if( pkt->pkttype != PKT_SIGNATURE && pkt->pkttype != PKT_MDC )
	        c->have_data = pkt->pkttype == PKT_PLAINTEXT;

	    if( newpkt == -1 )
	        ;
	    else if( newpkt ) {
	        pkt = xmalloc( sizeof *pkt );
	        init_packet(pkt);
	    }
	    else
	        free_packet(pkt);
	}
	if( rc == G10ERR_INVALID_PACKET )
	  write_status_text( STATUS_NODATA, "3" );
	if( any_data )
	  rc = 0;
	else if( rc == -1 )
	  write_status_text( STATUS_NODATA, "2" );


    leave:
	/* release_list( c ); */
	if (rc)
	    iobuf_cancel(c->out);
	else
	    iobuf_close(c->out);
	xfree(c->dek);
	free_packet( pkt );
	xfree( pkt );
	free_md_filter_context( &c->mfx );

        xfree( c );
    }
    /* rc = proc_encryption_packets( NULL, fp ); */
    if( no_out )
       opt.outfile = NULL;
    iobuf_close(fp);
    release_pk_list( pk_list );
    release_armor_context (afx);
    release_progress_context (pfx);
    return rc;
}
